"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _logger = require("../logger");

var _exceptions = require("../exceptions");

var _utils = require("../utils");

var _isomorphicWs = _interopRequireDefault(require("isomorphic-ws"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const config = require("../config");

class WebsocketClientBase {
  constructor(url, proxy, token) {
    this.url = url;
    this.proxy = proxy;
    this.token = token;
    this._connection = null;
    this._resolve = null;
    this._reject = null;
    this._error = null;
    this._messages = [];
    this._connectTimeout = null;
    this._pingInterval = null;
    this._pingTimeout = null;
  } // Override this method without calling super.

  /* abstract */


  connect() {
    throw new TypeError("Inheriting classes must implement connect");
  }

  isConnected() {
    return this._connection && this._connection.readyState === _isomorphicWs.default.OPEN;
  }

  read() {
    return new Promise((resolve, reject) => {
      try {
        if (this._error) {
          const error = this._error;
          this._error = null;
          reject(error);
        }

        const message = this._messages.shift();

        if (message) {
          resolve(message);
        }

        this.setPromiseHandlers(resolve, reject);
      } catch (e) {
        reject(e);
      }
    });
  } // Override this method without calling super.

  /* abstract */


  sendInner(message, timeout, resolve, reject) {
    throw new TypeError("Inheriting classes must implement sendInner");
  } // DON'T LOG HERE - THIS FUNCTION IS INDIRECTLY CALLED FROM WITHIN THE LOGGER
  // LOGGING HERE LEADS TO INFINITE RECURSION


  send(message) {
    return new Promise((resolve, reject) => {
      try {
        if (this._connection) {
          // For cases the callback was not called
          const timeout = setTimeout(() => {
            reject(new _exceptions.RookSendFailedError());
          }, config.AgentComConfiguration.SEND_RESOLVE_TIMEOUT);
          (0, _utils.nodeUnref)(timeout);
          this.sendInner(message, timeout, resolve, reject);
        } else {
          reject(new _exceptions.RookNotConnectedError());
        }
      } catch (e) {
        reject(e);
      }
    });
  } // Override this method without calling super.

  /* abstract */


  terminateConnection() {
    throw new TypeError("Inheriting classes must implement terminateConnection");
  }

  close() {
    this.terminateConnection();
    this.clearPingTimeout();
    this.clearPingInterval();
    this.clearConnectTimeout();
  } // Override this method without calling super.

  /* abstract */


  setResponseListeners() {
    throw new TypeError("Inheriting classes must implement setResponseListeners");
  }

  setPromiseHandlers(resolve, reject) {
    this._resolve = resolve;
    this._reject = reject;
  }

  returnConnected() {
    if (!this._resolve) {
      _logger.logger.warn("Failed to return connected status");
    }

    const resolve = this._resolve;
    this.clearPromiseHandlers();
    resolve();
  }

  returnError(error) {
    this._error = error;

    if (this._reject) {
      const reject = this._reject;
      this.clearPromiseHandlers();
      reject(error);
    }
  }

  returnMessage(message) {
    if (this._resolve) {
      const resolve = this._resolve;
      this.clearPromiseHandlers();
      resolve(message);
    } else {
      if (this._messages.length > config.AgentComConfiguration.MAX_QUEUED_MESSAGES) {
        _logger.logger.warn("Too many incoming messages, discarding");

        return;
      }

      this._messages.push(message);
    }
  }

  clearPromiseHandlers() {
    this._resolve = null;
    this._reject = null;
  }

  setConnectTimeout() {
    this.clearConnectTimeout();
    this._connectTimeout = setTimeout(() => {
      try {
        _logger.logger.error("WS connect timeout, closing connection");

        if (this._connection) {
          this.terminateConnection();
          this._connection = null;
        }
      } catch (e) {}
    }, config.AgentComConfiguration.CONNECT_TIMEOUT * 1000);
    (0, _utils.nodeUnref)(this._connectTimeout);
  }

  clearConnectTimeout() {
    if (this._connectTimeout) {
      clearTimeout(this._connectTimeout);
      this._connectTimeout = null;
    }
  }

  resetPingTimeout() {
    this.clearPingTimeout();
    this._pingTimeout = setTimeout(() => {
      try {
        _logger.logger.error("WS ping timeout, closing connection");

        if (this._connection) {
          this.terminateConnection();
          this._connection = null;
        }
      } catch (e) {}
    }, config.AgentComConfiguration.PING_TIMEOUT * 1000);
    (0, _utils.nodeUnref)(this._pingTimeout);
  }

  clearPingTimeout() {
    if (this._pingTimeout) {
      clearTimeout(this._pingTimeout);
      this._pingTimeout = null;
    }
  }

  setPingInterval() {
    if (this._pingInterval) {
      return;
    }

    this._pingInterval = setInterval(() => {
      try {
        if (this._connection && this._connection.readyState === _isomorphicWs.default.OPEN) {
          this._connection.ping(() => {});
        }
      } catch (e) {}
    }, config.AgentComConfiguration.PING_INTERVAL * 1000);
    (0, _utils.nodeUnref)(this._pingInterval);
  }

  clearPingInterval() {
    if (this._pingInterval) {
      clearInterval(this._pingInterval);
      this._pingInterval = null;
    }
  }

  heartbeat() {
    this.resetPingTimeout();
  }

}

exports.default = WebsocketClientBase;