'use strict';

var _utils = require("./utils");

var _exceptions = require("./exceptions");

var _sanitize = require("./sanitize");

const util = require('util');

const config = require('./config');

const {
  RookMissingToken,
  RookInvalidToken,
  RookInvalidOptions,
  RookOldServers
} = require('./exceptions');

const trueValues = ['y', 'Y', 'yes', 'Yes', 'YES', 'true', 'True', 'TRUE', '1', true];

class Rook {
  constructor() {
    this.savedOptions = null;
  }

  static _print_options(options = {}) {
    try {
      let config_string = "";

      for (let key in options) {
        if (options[key] !== undefined) {
          config_string = config_string + key + ": " + options[key] + ", ";
        }
      }

      console.log("RookOptions: " + config_string);
    } catch (e) {}
  }

  static _verifyBoolean(object, errorString) {
    if (!(typeof object === 'boolean')) {
      throw new RookInvalidOptions(errorString);
    }
  }

  static _verifyString(object, errorString) {
    if (!(typeof object === 'string' || object instanceof String)) {
      throw new RookInvalidOptions(errorString);
    }
  }

  static _verifyToken(object) {
    Rook._verifyString(object, 'Rookout token should be a String');

    if (object.length !== 64) {
      throw new RookInvalidOptions('Rookout token should be 64 characters');
    }

    if (/^[0-9a-zA-Z]+$/.test(object) === false) {
      throw new RookInvalidOptions('Rookout token must consist of only hexadecimal characters');
    }
  }

  static _verifyLabel(label) {
    if (label.startsWith("$")) {
      throw new _exceptions.RookInvalidLabel(label);
    }
  }

  static isVersionSupported() {
    if (!process || !process.versions || !process.versions.node) {
      return false;
    }

    let nodeVersion = process.versions.node;
    let version = nodeVersion.split(".");

    if (version.length < 3) {
      return false;
    }

    let processMajorVersion = parseInt(version[0]);
    let processMinorVersion = parseInt(version[1]);
    let processPatchVersion = parseInt(version[2]); // Validating not 8.9.1

    if (processMajorVersion === 8 && processMinorVersion === 9 && processPatchVersion === 1 || // Validating that 8.0.0 <= currentProcessVersion <= 12.22.12
    !(Rook.isVersionBiggerOrEqual(processMajorVersion, processMinorVersion, processPatchVersion, 8, 0, 0) && Rook.isVersionBiggerOrEqual(12, 22, 12, processMajorVersion, processMinorVersion, processPatchVersion)) && // Validating that 14.0.0 <= currentProcessVersion <= 14.21.3
    !(Rook.isVersionBiggerOrEqual(processMajorVersion, processMinorVersion, processPatchVersion, 14, 0, 0) && Rook.isVersionBiggerOrEqual(14, 21, 3, processMajorVersion, processMinorVersion, processPatchVersion)) && // Validating that 16.0.0 <= currentProcessVersion <= 16.19.1
    !(Rook.isVersionBiggerOrEqual(processMajorVersion, processMinorVersion, processPatchVersion, 16, 0, 0) && Rook.isVersionBiggerOrEqual(16, 19, 1, processMajorVersion, processMinorVersion, processPatchVersion)) && // Validating that 18.0.0 <= currentProcessVersion <= 18.15.0
    !(Rook.isVersionBiggerOrEqual(processMajorVersion, processMinorVersion, processPatchVersion, 18, 0, 0) && Rook.isVersionBiggerOrEqual(18, 15, 0, processMajorVersion, processMinorVersion, processPatchVersion))) {
      return false;
    }

    return true;
  }

  static isVersionDeprecated() {
    if (!process || !process.versions || !process.versions.node) {
      return false;
    }

    let nodeVersion = process.versions.node;
    let version = nodeVersion.split(".");

    if (version.length < 3) {
      return false;
    }

    let processMajorVersion = parseInt(version[0]); // Validating <= 8

    return processMajorVersion <= 8;
  }

  static isVersionBiggerOrEqual(biggerMajor, biggerMinor, biggerPatch, smallerMajor, smallerMinor, smallerPatch) {
    return biggerMajor > smallerMajor || biggerMajor === smallerMajor && (biggerMinor > smallerMinor || biggerMinor === smallerMinor && biggerPatch >= smallerPatch);
  }

  static parseSourcesList(sourcesList) {
    if (sourcesList === undefined) {
      if (!config.FrontendConfig.ENABLED) {
        let sources = process.env.ROOKOUT_SOURCES;

        if (sources !== undefined) {
          sourcesList = {};

          for (let source of sources.replace(/['"]/g, '').split(';')) {
            let kv = source.split('#');

            if (kv.length === 2) {
              let origin = kv[0];
              let revision = kv[1];

              if (origin && revision) {
                sourcesList[origin] = revision;
              }
            }
          }
        }
      }
    } else {
      // Normalize the map to native dictionary
      if (sourcesList instanceof Map) {
        let originalSources = sourcesList;
        sourcesList = {};

        for (let [origin, revision] of originalSources) {
          sourcesList[origin] = revision;
        }
      }

      for (let origin of Object.keys(sourcesList)) {
        Rook._verifyString(origin, 'Remote origin in source list should be a string');

        Rook._verifyString(sourcesList[origin], 'Revision in source list should be a string');
      }
    }

    return sourcesList;
  }

  envVarsUpdate(options) {
    if (options.quiet === undefined && process.env.ROOKOUT_QUIET !== undefined) {
      options.quiet = trueValues.indexOf(process.env.ROOKOUT_QUIET) !== -1;
    }

    if (options.log_to_stderr === undefined && process.env.ROOKOUT_LOG_TO_STDERR !== undefined) {
      options.log_to_stderr = trueValues.indexOf(process.env.ROOKOUT_LOG_TO_STDERR) !== -1;
    }

    if (options.ignore_node_modules === undefined && process.env.ROOKOUT_IGNORE_NODE_MODULES !== undefined) {
      options.ignore_node_modules = trueValues.indexOf(process.env.ROOKOUT_IGNORE_NODE_MODULES) !== -1;
    }

    if (options.sourcemap_debug === undefined && process.env.ROOKOUT_SOURCEMAP_DEBUG !== undefined) {
      options.sourcemap_debug = trueValues.indexOf(process.env.ROOKOUT_SOURCEMAP_DEBUG) !== -1;
    }

    if (options.log_file === undefined && process.env.ROOKOUT_LOG_FILE !== undefined) {
      options.log_file = process.env.ROOKOUT_LOG_FILE;
    }

    if (process.env.LAMBDA_TASK_ROOT) {
      options.log_file = "";

      if (options.lambda_safe_start !== true) {
        throw new _exceptions.RookIllegalLambdaStart();
      }
    }

    options.log_level = options.log_level || process.env.ROOKOUT_LOG_LEVEL;
  }

  resolveConfigSources(optionValue, envVarName, confValue) {
    if (config.FrontendConfig.ENABLED) {
      return optionValue || confValue;
    } else {
      return optionValue || process.env[envVarName] || confValue;
    }
  }

  start(options = {}) {
    if (Object.keys(options).length === 0 && this.savedOptions !== null) {
      options = this.savedOptions;
    }

    let self = this;
    this.startAttempted = true;
    (0, _sanitize.sanitize)(options);

    if (!config.FrontendConfig.ENABLED) {
      (0, _sanitize.sanitize)(process.env, true);
    }

    return new Promise((resolve, reject) => {
      try {
        if (self.singleton) resolve();
        self.debug = false;

        if (!config.FrontendConfig.ENABLED && options.debug === undefined) {
          self.debug = trueValues.indexOf(process.env.ROOKOUT_DEBUG) !== -1;
        } else {
          self.debug = options.debug || false;
        }

        self.envVarsUpdate(options);
        self.throw_errors = false;

        if (options.throw_errors !== undefined) {
          self.throw_errors = options.throw_errors;
        }

        if (!config.FrontendConfig.ENABLED) {
          const envSkipNodeWhitelist = trueValues.indexOf(process.env.ROOKOUT_SKIP_NODE_WHITELIST) !== -1;

          if (options.skip_node_whitelist || envSkipNodeWhitelist) {
            config.AgentComConfiguration.SKIP_NODE_VERSION_VERIFICATION = true;
          } else {
            config.AgentComConfiguration.SKIP_NODE_VERSION_VERIFICATION = false;
            config.DebuggerService.IS_SUPPORTED = Rook.isVersionSupported();
            config.DebuggerService.IS_DEPRECATED = Rook.isVersionDeprecated();
          }

          if (options.skip_ssl_verification === undefined && process.env.ROOKOUT_SKIP_SSL_VERIFY !== undefined) {
            config.AgentComConfiguration.SKIP_SSL_CERTIFICATE_VERIFICATION = trueValues.indexOf(process.env.ROOKOUT_SKIP_SSL_VERIFY) !== -1;
          }

          if (process.env.ROOKOUT_BREAKPOINT_FAILSAFE !== undefined) {
            config.InstrumentationConfig.EMPTY_HITBREAKPOINTS_FAILSAFE = trueValues.indexOf(process.env.ROOKOUT_BREAKPOINT_FAILSAFE);
          }
        }

        Rook._verifyBoolean(self.throw_errors, 'Rook throw errors should be a boolean');

        if (options.quick !== undefined) {
          Rook._verifyBoolean(options.quick, 'Rook quick should be a boolean');

          config.InstrumentationConfig.QUICK_START = options.quick;
        }

        if (options.quiet !== undefined) {
          Rook._verifyBoolean(options.quiet, 'Rook quiet should be a boolean');

          config.LoggingConfiguration.QUIET = options.quiet;
        }

        if (options.log_to_stderr !== undefined) {
          Rook._verifyBoolean(options.log_to_stderr, 'Rook log to stderr should be a boolean');

          config.LoggingConfiguration.LOG_TO_STDERR = options.log_to_stderr;
        }

        if (options.ignore_node_modules !== undefined) {
          Rook._verifyBoolean(options.ignore_node_modules, 'Rook ignore node modules should be a boolean');

          config.SourceFiles.IGNORE_NODE_MODULES = options.ignore_node_modules;
        }

        if (options.sourcemap_debug !== undefined) {
          Rook._verifyBoolean(options.sourcemap_debug, 'Rook sourcemap debug should be a boolean');

          config.LoggingConfiguration.ROOKOUT_SOURCEMAP_DEBUG = options.sourcemap_debug;
        }

        if (options.log_level !== undefined) {
          Rook._verifyString(options.log_level, 'Rook log level should be a String');

          config.LoggingConfiguration.LOG_LEVEL = options.log_level;
        }

        if (options.log_file !== undefined) {
          Rook._verifyString(options.log_file, 'Rook log file should be a String');

          config.LoggingConfiguration.FILE_NAME = options.log_file;
        }

        if (!config.FrontendConfig.ENABLED && options.tags === undefined) {
          let rawTags = process.env.ROOKOUT_ROOK_TAGS;

          if (rawTags !== undefined) {
            options.tags = [];

            for (let tag of rawTags.replace(/['"]/g, '').split(';')) {
              options.tags.push(_utils.StringUtils.trim(tag));
            }
          }
        } else if (options.tag !== undefined) {
          if (Array.isArray(options.tags)) {
            for (let tag of options.tags) {
              Rook._verifyString(tag, 'Rook tags should be array of strings');
            }
          } else {
            throw new RookInvalidOptions('Rook tags should be array of strings');
          }
        }

        if (!config.FrontendConfig.ENABLED && options.labels === undefined) {
          let labels = process.env.ROOKOUT_LABELS;

          if (labels !== undefined) {
            options.labels = {};

            for (let label of labels.replace(/['"]/g, '').split(',')) {
              let kv = label.split(':');

              if (kv.length === 2) {
                let key = kv[0];
                let value = kv[1];

                if (key && value) {
                  options.labels[key] = value;
                }
              }
            }
          }
        } else if (options.labels !== undefined) {
          // Normalize the map to native dictionary
          if (options.labels instanceof Map) {
            let originalLabels = options.labels;
            options.labels = {};

            for (let [key, value] of originalLabels) {
              options.labels[key] = value;
            }
          }

          for (let label of Object.keys(options.labels)) {
            // Set the label as an empty string if NULL
            if (options.labels[label] === undefined || options.labels[label] === null) {
              options.labels[label] = "";
            }

            Rook._verifyString(label);

            Rook._verifyString(options.labels[label], 'Rook label should be a map of strings');

            Rook._verifyLabel(label);
          }
        }

        Rook._verifyBoolean(self.debug, 'Rook debug flag should be a boolean');

        if (self.debug) {
          config.LoggingConfiguration.LOG_LEVEL = 'DEBUG';
          config.LoggingConfiguration.LOG_TO_STDERR = true;
          config.LoggingConfiguration.DEBUG = true;
        }

        if (options.git_commit !== undefined) {
          Rook._verifyString(options.git_commit, 'Git commit should be a String');

          config.GitConfiguration.GIT_COMMIT = options.git_commit;
        }

        if (options.git_origin !== undefined) {
          Rook._verifyString(options.git_origin, 'Git origin should be a String');

          config.GitConfiguration.GIT_ORIGIN = options.git_origin;
        }

        options.sources = Rook.parseSourcesList(options.sources);

        if (options.sources !== undefined) {
          config.GitConfiguration.SOURCES = options.sources;
        }

        config.InstrumentationConfig.MAX_SCRIPTS_COUNT_FAIL_SAFE = this.resolveConfigSources(options.maxScriptsFailSafe, "ROOKOUT_MAX_SCRIPTS_COUNT_FAIL_SAFE", config.InstrumentationConfig.MAX_SCRIPTS_COUNT_FAIL_SAFE);
        config.InstrumentationConfig.MAX_SCRIPTS_SIZE_FAIL_SAFE = this.resolveConfigSources(options.maxScriptsSizeFailSafe, "ROOKOUT_MAX_SCRIPTS_SIZE_FAIL_SAFE", config.InstrumentationConfig.MAX_SCRIPTS_SIZE_FAIL_SAFE);
        config.MonitorConfiguration.ENABLED = this.resolveConfigSources(options.enable_monitor, "ROOKOUT_ENABLE_MONITOR", config.MonitorConfiguration.ENABLED);
        config.MemoryUsageLogService.ENABLED = this.resolveConfigSources(options.enable_memory_usage_log, "ROOKOUT_ENABLE_MEMORY_USAGE_LOG", config.MemoryUsageLogService.ENABLED);
        config.MemoryUsageLogService.LOG_FILE = this.resolveConfigSources(options.memory_usage_log_file, "ROOKOUT_MEMORY_USAGE_LOG_FILE", config.MemoryUsageLogService.LOG_FILE);
        config.MemoryUsageLogService.TRACE_GC_ENABLED = this.resolveConfigSources(options.enable_trace_gc, "ROOKOUT_ENABLE_TRACE_GC", config.MemoryUsageLogService.TRACE_GC_ENABLED);
        let legacyLiveTailEnvVar = false;
        let liveTailEnvVar = false;
        let consoleLiveLoggerEnvVar = false;

        if (!config.FrontendConfig.ENABLED) {
          legacyLiveTailEnvVar = process.env.ROOKOUT_LIVE_TAIL;
          liveTailEnvVar = process.env.ROOKOUT_LIVE_LOGGER;
          consoleLiveLoggerEnvVar = options.console_live_logger || process.env.ROOKOUT_CONSOLE_LIVE_LOGGER;
        } else {
          // When using the rook for frontend, always enable console live logging
          consoleLiveLoggerEnvVar = options.console_live_logger;
        }

        let liveTailProp = options.live_logger;
        let legacyLiveTailProp = options.enable_livetail;
        let enableLiveTail = liveTailProp || legacyLiveTailProp || liveTailEnvVar || legacyLiveTailEnvVar || consoleLiveLoggerEnvVar || config.LogHandlerService.ENABLED || config.LogHandlerService.CONSOLE_FUNCTIONS_HOOK_ENABLED;
        config.LogHandlerService.ENABLED = enableLiveTail;

        if (enableLiveTail) {
          config.LogHandlerService.CONSOLE_FUNCTIONS_HOOK_ENABLED = consoleLiveLoggerEnvVar || config.LogHandlerService.CONSOLE_FUNCTIONS_HOOK_ENABLED;

          if (options.labels) {
            options.labels["live_logger"] = "on";
          } else {
            options.labels = {
              "live_logger": "on"
            };
          }
        }

        let heapSnapshotsDirectory = this.resolveConfigSources(options.heap_snapshots_directory, "ROOKOUT_HEAP_SNAPSHOTS_DIRECTORY", false);

        if (heapSnapshotsDirectory) {
          config.HeapSnapshotConfiguration.HEAP_PATH = heapSnapshotsDirectory;
          config.HeapSnapshotConfiguration.SIGNAL = this.resolveConfigSources(options.heap_snapshots_signal, "ROOKOUT_HEAP_SNAPSHOTS_SIGNAL", config.HeapSnapshotConfiguration.SIGNAL);
        }

        let host = options.host;
        let port = options.port;
        let proxy = options.proxy;
        let token = options.token;

        if (!config.FrontendConfig.ENABLED) {
          host = host || process.env.ROOKOUT_CONTROLLER_HOST || process.env.ROOKOUT_AGENT_HOST;
          port = port || process.env.ROOKOUT_CONTROLLER_PORT || process.env.ROOKOUT_AGENT_PORT;
          proxy = proxy || process.env.ROOKOUT_PROXY;
          token = token || process.env.ROOKOUT_TOKEN;
        }

        if (proxy !== undefined) {
          Rook._verifyString(proxy, 'proxy should be a string');
        }

        const {
          logger
        } = require('./logger');

        if (!config.FrontendConfig.ENABLED) {
          config.Tracing.ENABLED = require('./TracingPatcher').patchTracer();

          if (!config.Tracing.ENABLED) {
            logger.debug("Disabling tracing, hooking failed.");
          }
        } else {
          // No tracing in frontend
          config.Tracing.ENABLED = false;
        }

        let msg;

        if (config.FrontendConfig.ENABLED) {
          msg = `Rookout SDK ${config.VersionConfiguration.VERSION}:${config.VersionConfiguration.COMMIT} for frontend`;
        } else {
          msg = util.format('Rookout SDK [%s]:[%s] for NodeJS [%s]', config.VersionConfiguration.VERSION, config.VersionConfiguration.COMMIT, process.versions.node);
        }

        logger.debug(msg);

        if (config.InstrumentationConfig.QUICK_START) {
          logger.debug("Rookout started in quick mode");
        }

        if (host === undefined && token === undefined) {
          throw new RookMissingToken();
        } else {
          if (token !== undefined) {
            Rook._verifyToken(token);
          }
        }

        if (this.debug) {
          Rook._print_options({
            token: token !== undefined ? token.substring(0, 5) + "....." : undefined,
            host: host,
            port: port,
            proxy: proxy,
            throw_errors: self.throw_errors,
            log_level: options.log_level,
            log_to_stderr: options.log_to_stderr,
            log_file: options.log_file,
            git_commit: options.git_commit,
            git_origin: options.git_origin,
            sources: options.sources,
            tags: options.tags,
            live_logger: options.live_logger,
            labels: options.labels,
            quick: options.quick,
            quiet: options.quiet
          });
        }

        host = host || config.ControllerAddress.HOST;

        Rook._verifyString(host, 'Rook host should be String');

        if (host === "staging.cloud.agent.rookout.com" || host === "cloud.agent.rookout.com") {
          throw new RookOldServers();
        }

        port = port || config.ControllerAddress.PORT;

        if (!(typeof port === 'number' || typeof port === 'string')) {
          throw new RookInvalidOptions('Rook port should be a Number or a String');
        }

        if (!config.FrontendConfig.ENABLED && process.env.ROOKOUT_CONNECTION_TIMEOUT !== undefined) {
          if (typeof process.env.ROOKOUT_CONNECTION_TIMEOUT === 'number') {
            config.AgentComConfiguration.TIMEOUT = process.env.ROOKOUT_CONNECTION_TIMEOUT;
          }
        }

        this.savedOptions = options;
        self.singleton = require('./singleton').singleton;
        self.singleton.connect(token, host, port, proxy, options.tags, options.labels, !options.quick, this.debug, this.throw_errors).catch(e => {
          if (self.throw_errors) {
            reject(e);
          } else {
            if (e instanceof RookInvalidToken || e instanceof RookOldServers || e instanceof _exceptions.RookIllegalLambdaStart) {
              console.log('[Rookout] ', e.message);
            } else {
              logger.warn('[Rookout] Failed to connect to the controller - will continue attempting in the background: ', e.message);
            }

            if (self.debug) {
              console.error(e.stack || e);
            }

            resolve(self);
          }
        }).then(() => resolve(self));
      } catch (e) {
        if (self.throw_errors) {
          reject(e);
        } else {
          console.error('[Rookout] Failed to start Rookout: ', e.message);

          if (self.debug) {
            console.error(e.stack || e);
          }

          resolve(self);
        }
      }
    });
  }

  async stop() {
    if (!this.singleton) return;

    try {
      const singleton = this.singleton;
      this.singleton = undefined;
      await singleton.close();
    } catch (e) {
      if (this.throw_errors) {
        throw e;
      }

      if (this.debug) {
        console.error(e.stack || e);
      }
    }
  }

  flush(callback) {
    const retVal = new Promise(resolve => {
      if (!this.singleton) {
        resolve();
      }

      try {
        this.singleton.flush(resolve);
      } catch (e) {
        if (this.debug) {
          console.error(e.stack || e);
        }

        resolve();
      }
    });

    if (callback !== undefined) {
      retVal.then(callback);
    }

    return retVal;
  }

}

module.exports = Rook;