"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.updateVersionSupport = updateVersionSupport;

var _logger = require("./logger");

const process = require('process');

const config = require('./config');

function parseVersion(versionString) {
  const version = versionString.split('.');

  if (version.length < 3) {
    return undefined;
  }

  return version.map(x => parseInt(x));
}

function isVersionBiggerOrEqual(lhs, rhs) {
  return lhs[0] > rhs[0] || lhs[0] === rhs[0] && (lhs[1] > rhs[1] || lhs[1] === rhs[1] && lhs[2] >= rhs[2]);
}

function isInVersionRange(versionRange, ourVersion) {
  // The expected format is one of the following forms:
  // 1. "a.b.c-d.e.f" - where a-f are numbers (a.b.c is a semantic version), those are the versions in the range [a.b.c, d.e.f]
  // 2. "a.b.c-inf" - where we don't give an upper limit to the version range, those are the versions in the range [a.b.c, inf)
  const versions = versionRange.split('-');

  if (versions.length !== 2) {
    return false;
  }

  const rawLowerBoundVersion = versions[0];
  const rawUpperBoundVersion = versions[1];

  if (rawUpperBoundVersion === "inf") {
    const lowerBoundVersion = parseVersion(rawLowerBoundVersion);
    return lowerBoundVersion && isVersionBiggerOrEqual(ourVersion, lowerBoundVersion);
  } else {
    const lowerBoundVersion = parseVersion(rawLowerBoundVersion);
    const upperBoundVersion = parseVersion(rawUpperBoundVersion);
    return lowerBoundVersion && upperBoundVersion && isVersionBiggerOrEqual(ourVersion, lowerBoundVersion) && isVersionBiggerOrEqual(upperBoundVersion, ourVersion);
  }
}

function isVerifiedNodeVersion(nodeVersions) {
  const nodeVersionString = process.versions.node;
  const nodeVersion = parseVersion(process.versions.node);

  if (nodeVersion === undefined) {
    _logger.logger.debug("Failed to parse node version");

    return false;
  }

  for (const nodeVersionRange of nodeVersions) {
    if (isInVersionRange(nodeVersionRange, nodeVersion)) {
      return true;
    }
  }

  _logger.logger.debug(`Node version is not verified for this sdk version. SDK Version: ${"0.116.0"}, node version: ${nodeVersionString}`);

  return false;
}

function updateVersionSupport(rawVersions) {
  if (!config.FrontendConfig.ENABLED && !config.AgentComConfiguration.SKIP_NODE_VERSION_VERIFICATION) {
    var _process$versions;

    const parsedVersions = JSON.parse(rawVersions);
    const sdkVersion = parseVersion("0.116.0");

    if (sdkVersion === undefined) {
      _logger.logger.debug("Bad SDK Version");

      config.DebuggerService.IS_SUPPORTED = false;
      return;
    }

    if ((process === null || process === void 0 ? void 0 : (_process$versions = process.versions) === null || _process$versions === void 0 ? void 0 : _process$versions.node) === undefined) {
      _logger.logger.debug("Missing node version");

      config.DebuggerService.IS_SUPPORTED = false;
      return;
    } // Each key of the JSON object is an SDK version range (see `isInVersionRange` for format)


    for (const sdkVersionRange of Object.keys(parsedVersions)) {
      if (isInVersionRange(sdkVersionRange, sdkVersion)) {
        // Each value is a list of node version ranges
        const nodeVersions = parsedVersions[sdkVersionRange];
        const isSupported = isVerifiedNodeVersion(nodeVersions);

        if (isSupported) {
          config.DebuggerService.IS_SUPPORTED = true;
          return;
        }
      }
    }

    config.DebuggerService.IS_SUPPORTED = false;
  }
}